c      program pdWeintest
      subroutine pdWeintest
c     - swap between commenting out the above two lines to 
c     use this as a test, standalone programme, or as a routine
c     to be called.

c     proper distance, as per Weinberg (1972).

      parameter(mxcurv=5)
      real*4     qq(mxcurv),cc(mxcurv)
      real*4     q_0,H_0,cc_0
      parameter(mxred=4)
      real*4     red(mxred),prad(mxred),ptan(mxred)

      data  qq/-0.6, -0.55, -0.5, 0.15, 0.5  /
      data  cc/ 0.8,  0.7,  0.6,  0.0,  0.0  /
      data  red/1.0,3.0,30.0,1100.0/

      nc=mxcurv
      nred=4
      H_0=100.0
      do i=1,nc
         om=(qq(i)+cc(i))*2.0
         do iz=1,nred
            prad(iz)=pdWein(qq(i),H_0,cc(i), red(iz)) /1000.0
            ptan(iz)=arclen(qq(i),H_0,cc(i), red(iz)) /1000.0
         enddo
         
         write(6,'(2f7.2,5f7.1,f7.0)')
     _        om,cc(i),prad(1),prad(2),prad(3),ptan(3),prad(4),
     _        prad(3)/prad(4)*100.0

         print*,'rC=',rcurvfn(qq(i),H_0,cc(i))
      enddo

      do while(1+1.eq.2)
      write(6,'(a,$)')'Enter q_0, H_0, cc_0, z: '
      read(5,*)q_0,H_0,cc_0, z
      print*,q_0,H_0,cc_0, z

      pdW= pdWein(q_0,H_0,cc_0,z)
      print*,'pdW= ',pdW
      dpm2= propd2(q_0,H_0,cc_0,z)
      print*,'dpm2= ',dpm2

      zW= zofpdW(q_0,H_0,cc_0, pdW)
      print*,' zW=', zW

      a= arclen(q_0,H_0,cc_0, z)
      print*,'arclen  a=',a
      zarc= zofarc(q_0,H_0,cc_0, a)
      print*,'zofarc  zarc=',zarc

      enddo

      end


      real*4 function pdWein(q_0,H_0,cc_0,z)
c     proper distance, as per Weinberg (1972).

      real*8     kapp0,om0,cc_0d, factor, d, zero, c
      real*8     rcurv

      logical    verbose

      parameter(c=299792.4562)
      parameter(zero=  1.0d-5)
      
      verbose=.false.

      cc_0d=dble(cc_0)
      om0= 2.0d0*(dble(q_0)+cc_0d)
      kapp0= om0 + cc_0d -1.0d0

      d= dble(propd2(q_0,H_0,cc_0,z))

      if(abs(kapp0).lt.zero)then
         pdWein= real(d)
         return
c         factor=1.0d0
      else
         factor= sqrt(abs(kapp0))
      endif
      if(verbose)print*,'pdWein: om0,kapp0,factor=',om0,kapp0,factor
      if(verbose)print*,'cc_0d=',cc_0d
      
      rcurv= c/factor /dble(H_0)
      if(verbose)print*,'Rcurv=',rcurv
      
      d= d / rcurv

      if(kapp0.le.-zero)then
         d= log( d + sqrt(d*d +1.0d0) ) ! d= asinh( d )
      elseif(kapp0.gt.zero)then
         if(d.lt.zero.or.d.gt.1.0d0+zero)then
            print*,'pdWein Warning: d =',d,' out of range!'
            d=max(min(0.99999d0,d),0.00001d0)
         endif
         if(d.gt.1.0d0)d= min(0.99999d0,d)
         d= asin( d )
      endif
      d= rcurv * d
      
      pdWein= real(d)
      return
      end
      

      real*4 function zofpdW(q_0,H_0,cc_0,dp)
c     ==== for a given proper distance (Weinberg 1972), find z ====
c     i.e., inverse of  pdWein(...)

      real*4 q_0,H_0,cc_0,dp !inputs

      real*8     kapp0,om0,cc_0d, factor, d, zero, c !internal
      real*8     rcurv          !internal
      real*4     d4

      logical    verbose

      parameter(c=299792.4562)
      parameter(zero=  1.0d-5)
      
      verbose=.false.

      cc_0d=dble(cc_0)
      om0= 2.0d0*(dble(q_0)+cc_0d)
      kapp0= om0 + cc_0d -1.0d0

      if(abs(kapp0).lt.zero)then
         zofpdW= zofpd2(q_0,H_0,cc_0, dp)
         return
c         factor=1.0d0
      else
         factor= sqrt(abs(kapp0))
      endif
      if(verbose)print*,'zofpdW: om0,kapp0,factor=',om0,kapp0,factor
      if(verbose)print*,'cc_0d=',cc_0d
      
      rcurv= c/factor /dble(H_0)
      if(verbose)print*,'Rcurv=',rcurv
      
      d= dble(dp) /rcurv ! proper dist in units of R_C

c     -- convert to proper motion distance, i.e. propd2( )
      if(kapp0.le.-zero)then
         d= (exp(d) -exp(-d) )/2.0d0 ! d= sinh( d )
      elseif(kapp0.gt.zero)then
         d= sin( d )
      endif
      d= rcurv * d
      d4=real(d)

      zofpdW= zofpd2(q_0,H_0,cc_0, d4)
      return
      end


      real*4 function zofpdW_old(q_0,H_0,cc_0,dp)
c     ==== for a given proper distance (Weinberg 1972), find z ====
      real*4 q_0,H_0,cc_0,dp

      real*4 z1,z2, z3
      logical*1  verbose
      verbose=.false.

      z3= zofpd2(q_0,H_0,cc_0, dp)

      z1= 0.5*z3
      z2= 2.0*z3

      tol= 1e-4
      maxiter=30
      niter=0

      d1= pdWein(q_0,H_0,cc_0, z1)
      d2= pdWein(q_0,H_0,cc_0, z2)

      do while(z2-z1.gt.tol.and.niter.lt.maxiter)
         if(dp.lt.d1)then
            z1=0.5*z1
         elseif(dp.gt.d2)then
            z2=2.0*z2
         else
            z3= (z1+z2)/2.0
            d3= pdWein(q_0,H_0,cc_0, z3)
            if(d3.lt.dp)then
               z1=z3
            else
               z2=z3
            endif
         endif
         d1= pdWein(q_0,H_0,cc_0, z1)
         d2= pdWein(q_0,H_0,cc_0, z2)
         niter=niter+1
      enddo
      
      if(verbose)print*,'niter, dp, z1,z2,d1,d2=',
     _     niter, dp, z1,z2,d1,d2

      zofpdW= (z1+z2)/2.0
      return
      end


      function arclen(q0,H0,cc0,z)
c     --- length along arc of 1 radian in comoving, proper distance units ---
      dd= pdWein(q0,H0,cc0, z)
      tol=1e-5
      curv0= 2.0*q0+3.0*cc0 -1.0
      if(curv0.lt.-tol)then !  hyperbolic
         rC= 3000.0 *(100.0/H0) /sqrt( - curv0 )
         arclen= rC* (exp(dd/rC) - exp(-dd/rC))/2.0
      elseif(curv0.gt.-tol .and. curv0.lt.tol)then
         arclen =  dd 
      else
         rC= 3000.0 *(100.0/H0)  /sqrt( curv0 )
         arclen = rC*sin(dd/rC)
      endif
      return
      end


      function rcurvfn(q0,H0,cc0)
c     --- curvature radius comoving, proper distance units ---
      tol=1e-5
      curv0= 2.0*q0+3.0*cc0 -1.0
      if(curv0.lt.-tol)then !  hyperbolic
         rC= 3000.0 *(100.0/H0) /sqrt( - curv0 )
      elseif(curv0.gt.-tol .and. curv0.lt.tol)then
         rC=-99.9
      else
         rC= 3000.0 *(100.0/H0)  /sqrt( curv0 )
      endif
      rcurvfn=rC
      return
      end


      function zofarc(q0,H0,cc0,arc)
c     --- redshift at which arclength  arc = 1 radian in comov. prop dist ---
      tol=1e-5
      tol_sin= 1e-5

      curv0= 2.0*q0+3.0*cc0 -1.0
      if(curv0.lt.-tol)then !  hyperbolic      
         rC= 3000.0 *(100.0/H0) /sqrt( - curv0 )
         arcC = arc/rC
         dd= rC * log(arcC + sqrt(arcC*arcC +1.0))
         zofarc = zofpdW(q0,H0,cc0, dd)
      elseif(curv0.gt.-tol .and. curv0.lt.tol)then
         zofarc = zofpdW(q0,H0,cc0, arc)
      else
c     !       arc/rC < pi/2 ! !
         rC= 3000.0 *(100.0/H0) /sqrt( curv0 )
         arcC = arc/rC
         if(arcC.gt.1.0)then
            print*,'zofarc Warning: arcC too big ! '
            print*,'q0,H0,cc0,arc, rC =',q0,H0,cc0,arc, rC 
            arcC = 1.0 - tol_sin
         endif
         dd = rC * asin(arcC)
         zofarc = zofpdW(q0,H0,cc0, dd)
      endif
      return
      end
